"""
Helpers and examples for supported API Requests that each Celery
Task supports:

- analysis_engine.work_tasks.get_new_pricing_data
- analysis_engine.work_tasks.handle_pricing_update_task
- analysis_engine.work_tasks.publish_pricing_update

"""

import datetime
import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.iex.consts as iex_consts
import analysis_engine.td.consts as td_consts
import analysis_engine.options_dates as opt_dates


def get_ds_dict(
        ticker,
        base_key=None,
        ds_id=None,
        label=None,
        service_dict=None):
    """get_ds_dict

    Get a dictionary with all cache keys for a ticker and return
    the dictionary. Use this method to decouple your apps
    from the underlying cache key implementations (if you
    do not need them).

    :param ticker: ticker
    :param base_key: optional - base key that is prepended
                     in all cache keys
    :param ds_id: optional - dataset id (useful for
                  external database id)
    :param label: optional - tracking label in the logs
    :param service_dict: optional - parent call functions and Celery
                         tasks can use this dictionary to seed the
                         common service routes and endpoints. Refer
                         to ``analysis_engine.consts.SERVICE_VALS``
                         for automatically-copied over keys by this
                         helper.
    """

    if not ticker:
        raise Exception('please pass in a ticker')

    use_base_key = base_key
    if not use_base_key:
        last_str = ae_utils.get_last_close_str(
            fmt=ae_consts.COMMON_DATE_FORMAT)
        use_base_key = f'{ticker}_{last_str}'

    date_str = ae_utils.utc_date_str(
        fmt=ae_consts.COMMON_DATE_FORMAT)
    now_str = ae_utils.utc_now_str(
        fmt=ae_consts.COMMON_TICK_DATE_FORMAT)

    daily_redis_key = (
        f'{use_base_key}_{ae_consts.DAILY_S3_BUCKET_NAME}')
    minute_redis_key = (
        f'{use_base_key}_{ae_consts.MINUTE_S3_BUCKET_NAME}')
    quote_redis_key = (
        f'{use_base_key}_{ae_consts.QUOTE_S3_BUCKET_NAME}')
    stats_redis_key = (
        f'{use_base_key}_{ae_consts.STATS_S3_BUCKET_NAME}')
    peers_redis_key = (
        f'{use_base_key}_{ae_consts.PEERS_S3_BUCKET_NAME}')
    news_iex_redis_key = (
        f'{use_base_key}_{ae_consts.NEWS_S3_BUCKET_NAME}1')
    financials_redis_key = (
        f'{use_base_key}_{ae_consts.FINANCIALS_S3_BUCKET_NAME}')
    earnings_redis_key = (
        f'{use_base_key}_{ae_consts.EARNINGS_S3_BUCKET_NAME}')
    dividends_redis_key = (
        f'{use_base_key}_{ae_consts.DIVIDENDS_S3_BUCKET_NAME}')
    company_redis_key = (
        f'{use_base_key}_{ae_consts.COMPANY_S3_BUCKET_NAME}')
    options_yahoo_redis_key = (
        f'{use_base_key}_{ae_consts.OPTIONS_S3_BUCKET_NAME}')
    call_options_yahoo_redis_key = (
        f'{use_base_key}_calls')
    put_options_yahoo_redis_key = (
        f'{use_base_key}_puts')
    pricing_yahoo_redis_key = (
        f'{use_base_key}_{ae_consts.PRICING_S3_BUCKET_NAME}')
    news_yahoo_redis_key = (
        f'{use_base_key}_{ae_consts.NEWS_S3_BUCKET_NAME}')
    call_options_td_redis_key = (
        f'{use_base_key}_tdcalls')
    put_options_td_redis_key = (
        f'{use_base_key}_tdputs')

    ds_cache_dict = {
        'daily': daily_redis_key,
        'minute': minute_redis_key,
        'quote': quote_redis_key,
        'stats': stats_redis_key,
        'peers': peers_redis_key,
        'news1': news_iex_redis_key,
        'financials': financials_redis_key,
        'earnings': earnings_redis_key,
        'dividends': dividends_redis_key,
        'company': company_redis_key,
        'options': options_yahoo_redis_key,
        'calls': call_options_yahoo_redis_key,
        'puts': put_options_yahoo_redis_key,
        'pricing': pricing_yahoo_redis_key,
        'news': news_yahoo_redis_key,
        'tdcalls': call_options_td_redis_key,
        'tdputs': put_options_td_redis_key,
        'ticker': ticker,
        'ds_id': ds_id,
        'label': label,
        'created': now_str,
        'date': date_str,
        'manifest_key': use_base_key,
        'version': ae_consts.CACHE_DICT_VERSION
    }

    # set keys/values for redis/minio from the
    # service_dict - helper method for
    # launching job chains
    if service_dict:
        for k in ae_consts.SERVICE_VALS:
            ds_cache_dict[k] = service_dict[k]

    return ds_cache_dict
# end of get_ds_dict


def build_get_new_pricing_request(
        label=None):
    """build_get_new_pricing_request

    Build a sample Celery task API request:
    analysis_engine.work_tasks.get_new_pricing_data

    Used for testing: run_get_new_pricing_data(work)

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    base_key = '{}_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.S3_BUCKET
    s3_key = base_key
    redis_key = base_key
    use_strike = None
    contract_type = 'C'
    get_pricing = True
    get_news = True
    get_options = True

    work = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        'strike': use_strike,
        'contract': contract_type,
        'get_pricing': get_pricing,
        'get_news': get_news,
        'get_options': get_options
    }

    if label:
        work['label'] = label

    return work
# end of build_get_new_pricing_request


def build_cache_ready_pricing_dataset(
        label=None):
    """build_cache_ready_pricing_dataset

    Build a cache-ready pricing dataset to replicate
    the ``get_new_pricing_data`` task

    :param label: log label to use
    """

    pricing_dict = {
        'ask': 0.0,
        'askSize': 8,
        'averageDailyVolume10Day': 67116380,
        'averageDailyVolume3Month': 64572187,
        'bid': 0.0,
        'bidSize': 10,
        'close': 287.6,
        'currency': 'USD',
        'esgPopulated': False,
        'exchange': 'PCX',
        'exchangeDataDelayedBy': 0,
        'exchangeTimezoneName': 'America/New_York',
        'exchangeTimezoneShortName': 'EDT',
        'fiftyDayAverage': 285.21735,
        'fiftyDayAverageChange': 2.8726501,
        'fiftyDayAverageChangePercent': 0.010071794,
        'fiftyTwoWeekHigh': 291.74,
        'fiftyTwoWeekHighChange': -3.649994,
        'fiftyTwoWeekHighChangePercent': -0.012511119,
        'fiftyTwoWeekLow': 248.02,
        'fiftyTwoWeekLowChange': 40.069992,
        'fiftyTwoWeekLowChangePercent': 0.16155952,
        'fiftyTwoWeekRange': '248.02 - 291.74',
        'financialCurrency': 'USD',
        'fullExchangeName': 'NYSEArca',
        'gmtOffSetMilliseconds': -14400000,
        'high': 289.03,
        'language': 'en-US',
        'longName': 'SPDR S&amp;P 500 ETF',
        'low': 287.88,
        'market': 'us_market',
        'marketCap': 272023797760,
        'marketState': 'POSTPOST',
        'messageBoardId': 'finmb_6160262',
        'open': 288.74,
        'postMarketChange': 0.19998169,
        'postMarketChangePercent': 0.06941398,
        'postMarketPrice': 288.3,
        'postMarketTime': 1536623987,
        'priceHint': 2,
        'quoteSourceName': 'Delayed Quote',
        'quoteType': 'ETF',
        'region': 'US',
        'regularMarketChange': 0.48999023,
        'regularMarketChangePercent': 0.17037213,
        'regularMarketDayHigh': 289.03,
        'regularMarketDayLow': 287.88,
        'regularMarketDayRange': '287.88 - 289.03',
        'regularMarketOpen': 288.74,
        'regularMarketPreviousClose': 287.6,
        'regularMarketPrice': 288.09,
        'regularMarketTime': 1536609602,
        'regularMarketVolume': 50210903,
        'sharesOutstanding': 944232000,
        'shortName': 'SPDR S&P 500',
        'sourceInterval': 15,
        'symbol': 'SPY',
        'tradeable': True,
        'trailingThreeMonthNavReturns': 7.71,
        'trailingThreeMonthReturns': 7.63,
        'twoHundredDayAverage': 274.66153,
        'twoHundredDayAverageChange': 13.428467,
        'twoHundredDayAverageChangePercent': 0.048890963,
        'volume': 50210903,
        'ytdReturn': 9.84
    }
    calls_df_as_json = pd.DataFrame([{
        'ask': 106,
        'bid': 105.36,
        'change': 4.0899963,
        'contractSize': 'REGULAR',
        'contractSymbol': 'SPY181019P00380000',
        'currency': 'USD',
        'expiration': 1539907200,
        'impliedVolatility': 1.5991230981,
        'inTheMoney': True,
        'lastPrice': 91.82,
        'lastTradeDate': 1539027901,
        'openInterest': 0,
        'percentChange': 4.4543633,
        'strike': 380,
        'volume': 37
    }]).to_json(
        orient='records')
    puts_df_as_json = pd.DataFrame([{
        'ask': 106,
        'bid': 105.36,
        'change': 4.0899963,
        'contractSize': 'REGULAR',
        'contractSymbol': 'SPY181019P00380000',
        'currency': 'USD',
        'expiration': 1539907200,
        'impliedVolatility': 1.5991230981,
        'inTheMoney': True,
        'lastPrice': 91.82,
        'lastTradeDate': 1539027901,
        'openInterest': 0,
        'percentChange': 4.4543633,
        'strike': 380,
        'volume': 37
    }]).to_json(
        orient='records')
    news_list = [
        {
            'd': '16 hours ago',
            's': 'Yahoo Finance',
            'sp': 'Some Title 1',
            'sru': 'http://news.google.com/news/url?values',
            't': 'Some Title 1',
            'tt': '1493311950',
            'u': 'http://finance.yahoo.com/news/url1',
            'usg': 'ke1'
        },
        {
            'd': '18 hours ago',
            's': 'Yahoo Finance',
            'sp': 'Some Title 2',
            'sru': 'http://news.google.com/news/url?values',
            't': 'Some Title 2',
            'tt': '1493311950',
            'u': 'http://finance.yahoo.com/news/url2',
            'usg': 'key2'
        }
    ]

    options_dict = {
        'exp_date': '2018-10-19',
        'calls': calls_df_as_json,
        'puts': puts_df_as_json,
        'num_calls': 1,
        'num_puts': 1
    }

    cache_data = {
        'news': news_list,
        'options': options_dict,
        'pricing': pricing_dict
    }
    return cache_data
# end of build_cache_ready_pricing_dataset


def build_publish_pricing_request(
        label=None):
    """build_publish_pricing_request

    Build a sample Celery task API request:
    analysis_engine.work_tasks.publisher_pricing_update

    Used for testing: run_publish_pricing_update(work)

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    base_key = '{}_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.S3_BUCKET
    s3_key = base_key
    redis_key = base_key
    use_strike = None
    contract_type = 'C'
    use_data = build_cache_ready_pricing_dataset()

    work = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        'strike': use_strike,
        'contract': contract_type,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        'df_compress': True,
        'data': use_data
    }

    if label:
        work['label'] = label

    return work
# end of build_publish_pricing_request


def build_publish_from_s3_to_redis_request(
        label=None):
    """build_publish_from_s3_to_redis_request

    Build a sample Celery task API request:
    analysis_engine.work_tasks.publish_from_s3_to_redis

    Used for testing: run_publish_from_s3_to_redis(work)

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    base_key = '{}_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.S3_BUCKET
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_publish_from_s3_to_redis_request


def build_publish_ticker_aggregate_from_s3_request(
        label=None):
    """build_publish_ticker_aggregate_from_s3_request

    Build a sample Celery task API request:
    analysis_engine.work_tasks.publish_ticker_aggregate_from_s3

    Used for testing: run_publish_ticker_aggregate_from_s3(work)

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    s3_bucket_name = ae_consts.S3_BUCKET
    s3_compiled_bucket_name = ae_consts.S3_COMPILED_BUCKET
    s3_key = '{}_latest'.format(ticker)
    redis_key = '{}_latest'.format(ticker)
    s3_enabled = True
    redis_enabled = True

    work = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        's3_bucket': s3_bucket_name,
        's3_compiled_bucket': s3_compiled_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_publish_ticker_aggregate_from_s3_request


def build_prepare_dataset_request(
        label=None):
    """build_prepare_dataset_request

    Build a sample Celery task API request:
    analysis_engine.work_tasks.prepare_pricing_dataset

    Used for testing: run_prepare_pricing_dataset(work)

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    base_key = '{}_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.S3_BUCKET
    s3_key = base_key
    redis_key = base_key
    s3_prepared_bucket_name = ae_consts.PREPARE_S3_BUCKET_NAME
    s3_prepared_key = '{}.csv'.format(
        base_key)
    redis_prepared_key = '{}'.format(
        base_key)
    ignore_columns = None
    s3_enabled = True
    redis_enabled = True

    work = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        'prepared_s3_key': s3_prepared_key,
        'prepared_s3_bucket': s3_prepared_bucket_name,
        'prepared_redis_key': redis_prepared_key,
        'ignore_columns': ignore_columns,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_prepare_dataset_request


def build_analyze_dataset_request(
        label=None):
    """build_analyze_dataset_request

    Build a sample Celery task API request:
    analysis_engine.work_tasks.analyze_pricing_dataset

    Used for testing: run_analyze_pricing_dataset(work)

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    base_key = '{}_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.PREPARE_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_analyzed_bucket_name = ae_consts.ANALYZE_S3_BUCKET_NAME
    s3_analyzed_key = '{}.csv'.format(
        base_key)
    redis_analyzed_key = '{}'.format(
        base_key)
    ignore_columns = None
    s3_enabled = True
    redis_enabled = True

    work = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        'analyzed_s3_key': s3_analyzed_key,
        'analyzed_s3_bucket': s3_analyzed_bucket_name,
        'analyzed_redis_key': redis_analyzed_key,
        'ignore_columns': ignore_columns,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_analyze_dataset_request


"""
IEX API Requests
"""


def build_iex_fetch_daily_request(
        label=None):
    """build_iex_fetch_daily_request

    Fetch `daily
    data <https://iexcloud.io/docs/api/#historical-prices>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_daily_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.DAILY_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_DAILY,
        'fd_type': iex_consts.DATAFEED_DAILY,
        'ticker': ticker,
        'timeframe': '5y',
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_daily_request


def build_iex_fetch_minute_request(
        label=None):
    """build_iex_fetch_minute_request

    Fetch `minute
    data <https://iexcloud.io/docs/api/#historical-prices>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_minute_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.MINUTE_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_MINUTE,
        'fd_type': iex_consts.DATAFEED_MINUTE,
        'ticker': ticker,
        'timeframe': '1d',
        'last_close': None,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_minute_request


def build_iex_fetch_quote_request(
        label=None):
    """build_iex_fetch_quote_request

    Fetch `quote
    data <https://iexcloud.io/docs/api/#quote>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_quote_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.QUOTE_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_QUOTE,
        'fd_type': iex_consts.DATAFEED_QUOTE,
        'ticker': ticker,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_quote_request


def build_iex_fetch_stats_request(
        label=None):
    """build_iex_fetch_stats_request

    Fetch `stats
    data <https://iexcloud.io/docs/api/#key-stats>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_stat_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.STATS_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_STATS,
        'fd_type': iex_consts.DATAFEED_STATS,
        'ticker': ticker,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_stats_request


def build_iex_fetch_peers_request(
        label=None):
    """build_iex_fetch_peers_request

    Fetch `peers
    data <https://iexcloud.io/docs/api/#peers>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_peer_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.PEERS_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_PEERS,
        'fd_type': iex_consts.DATAFEED_PEERS,
        'ticker': ticker,
        'timeframe': '1d',
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_peers_request


def build_iex_fetch_news_request(
        label=None):
    """build_iex_fetch_news_request

    Fetch `news
    data <https://iexcloud.io/docs/api/#news>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_news_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.NEWS_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_NEWS,
        'fd_type': iex_consts.DATAFEED_NEWS,
        'ticker': ticker,
        'timeframe': '1d',
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_news_request


def build_iex_fetch_financials_request(
        label=None):
    """build_iex_fetch_financials_request

    Fetch `financials
    data <https://iexcloud.io/docs/api/#financials>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_financial_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.FINANCIALS_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_FINANCIALS,
        'fd_type': iex_consts.DATAFEED_FINANCIALS,
        'ticker': ticker,
        'timeframe': '1d',
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_financials_request


def build_iex_fetch_earnings_request(
        label=None):
    """build_iex_fetch_earnings_request

    Fetch `earnings
    data <https://iexcloud.io/docs/api/#earnings>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_earning_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.EARNINGS_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_EARNINGS,
        'fd_type': iex_consts.DATAFEED_EARNINGS,
        'ticker': ticker,
        'timeframe': '1d',
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_earnings_request


def build_iex_fetch_dividends_request(
        label=None):
    """build_iex_fetch_dividends_request

    Fetch `dividends
    data <https://iexcloud.io/docs/api/#dividends>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_dividend_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.DIVIDENDS_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_DIVIDENDS,
        'fd_type': iex_consts.DATAFEED_DIVIDENDS,
        'ticker': ticker,
        'timeframe': '2y',
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_dividends_request


def build_iex_fetch_company_request(
        label=None):
    """build_iex_fetch_company_request

    Fetch `company
    data <https://iexcloud.io/docs/api/#company>`__
    from IEX

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_company_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = ae_consts.COMPANY_S3_BUCKET_NAME
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'ft_type': iex_consts.FETCH_COMPANY,
        'fd_type': iex_consts.DATAFEED_COMPANY,
        'ticker': ticker,
        'timeframe': '1d',
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_iex_fetch_company_request


def build_screener_analysis_request(
        ticker=None,
        tickers=None,
        fv_urls=None,
        fetch_mode='iex',
        iex_datasets=ae_consts.IEX_DATASETS_DEFAULT,
        determine_sells=None,
        determine_buys=None,
        label='screener'):
    """build_screener_analysis_request

    Build a dictionary request for the task:
    ``analysis_engine.work_tasks.run_screener_analysis``

    :param ticker: ticker to add to the analysis
    :param tickers: tickers to add to the analysis
    :param fv_urls: finviz urls
    :param fetch_mode: supports pulling from ``iex``,
        ``yahoo``, ``all`` (defaults to ``iex``)
    :param iex_datasets: datasets to fetch from
        ``iex`` (defaults to ``analysis_engine.con
        sts.IEX_DATASETS_DEFAULT``)
    :param determine_sells: string custom Celery task
        name for handling sell-side processing
    :param determine_buys: string custom Celery task
        name for handling buy-side processing
    :param label: log tracking label
    :return: initial request dictionary:
        ::

            req = {
                'tickers': use_tickers,
                'fv_urls': use_urls,
                'fetch_mode': fetch_mode,
                'iex_datasets': iex_datasets,
                's3_bucket': s3_bucket_name,
                's3_enabled': s3_enabled,
                'redis_enabled': redis_enabled,
                'determine_sells': determine_sells,
                'determine_buys': determine_buys,
                'label': label
            }
    """
    use_urls = []
    if fv_urls:
        for f in fv_urls:
            if f not in use_urls:
                use_urls.append(f)

    use_tickers = tickers
    if ticker:
        if not tickers:
            use_tickers = [
                ticker
            ]
        if ticker.upper() not in use_tickers:
            use_tickers.append(ticker.upper())

    s3_bucket_name = ae_consts.SCREENER_S3_BUCKET_NAME
    s3_enabled = True
    redis_enabled = True

    req = {
        'tickers': use_tickers,
        'urls': use_urls,
        'fetch_mode': fetch_mode,
        'iex_datasets': iex_datasets,
        's3_bucket': s3_bucket_name,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled,
        'determine_sells': determine_sells,
        'determine_buys': determine_buys,
        'label': label
    }
    return req
# end build_screener_analysis_request


"""
Tradier API Requests
"""


def build_td_fetch_calls_request(
        label=None):
    """build_td_fetch_calls_request

    Fetch tradier calls

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_tdcalls_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = 'tdcalls'
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    exp_date = opt_dates.option_expiration().strftime(
        ae_consts.COMMON_DATE_FORMAT)

    work = {
        'ft_type': td_consts.FETCH_TD_CALLS,
        'fd_type': td_consts.DATAFEED_TD_CALLS,
        'ticker': ticker,
        'exp_date': exp_date,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_td_fetch_calls_request


def build_td_fetch_puts_request(
        label=None):
    """build_td_fetch_puts_request

    Fetch tradier puts

    :param label: log label to use
    """
    ticker = ae_consts.TICKER
    base_key = '{}_tdputs_{}'.format(
        ticker,
        datetime.datetime.utcnow().strftime(
            '%Y_%m_%d_%H_%M_%S'))
    s3_bucket_name = 'tdputs'
    s3_key = base_key
    redis_key = base_key
    s3_enabled = True
    redis_enabled = True

    exp_date = opt_dates.option_expiration().strftime(
        ae_consts.COMMON_DATE_FORMAT)

    work = {
        'ft_type': td_consts.FETCH_TD_PUTS,
        'fd_type': td_consts.DATAFEED_TD_PUTS,
        'ticker': ticker,
        'exp_date': exp_date,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

    if label:
        work['label'] = label

    return work
# end of build_td_fetch_puts_request



"""
Parse data from yahoo
"""

import copy
import datetime
import pinance
import analysis_engine.options_dates as opt_dates
import analysis_engine.get_pricing as yahoo_get_pricing
import analysis_engine.build_result as build_result
import analysis_engine.work_tasks.publish_pricing_update as \
    publisher
import spylunking.log.setup_logging as log_utils
from analysis_engine.consts import TICKER
from analysis_engine.consts import SUCCESS
from analysis_engine.consts import NOT_RUN
from analysis_engine.consts import ERR
from analysis_engine.consts import NOT_SET
from analysis_engine.consts import COMMON_TICK_DATE_FORMAT
from analysis_engine.consts import EMPTY_DF_STR
from analysis_engine.consts import get_status
from analysis_engine.consts import ppj
from analysis_engine.utils import get_last_close_str

log = log_utils.build_colorized_logger(
    name=__name__)


def get_data_from_yahoo(
        work_dict):
    """get_data_from_yahoo

    Get data from yahoo

    :param work_dict: request dictionary
    """
    label = 'get_data_from_yahoo'

    log.info(
        'task - {} - start '
        'work_dict={}'.format(
            label,
            work_dict))

    num_news_rec = 0
    num_option_calls = 0
    num_option_puts = 0
    cur_high = -1
    cur_low = -1
    cur_open = -1
    cur_close = -1
    cur_volume = -1

    rec = {
        'pricing': None,
        'options': None,
        'calls': None,
        'puts': None,
        'news': None,
        'exp_date': None,
        'publish_pricing_update': None,
        'date': None,
        'updated': None
    }
    res = {
        'status': NOT_RUN,
        'err': None,
        'rec': rec
    }

    try:

        ticker = work_dict.get(
            'ticker',
            TICKER)
        exp_date = work_dict.get(
            'exp_date',
            None)
        cur_strike = work_dict.get(
            'strike',
            None)
        contract_type = str(work_dict.get(
            'contract',
            'C')).upper()
        get_pricing = work_dict.get(
            'get_pricing',
            True)
        get_news = work_dict.get(
            'get_news',
            True)
        get_options = work_dict.get(
            'get_options',
            True)
        orient = work_dict.get(
            'orient',
            'records')
        label = work_dict.get(
            'label',
            label)

        ticker_results = pinance.Pinance(ticker)
        num_news_rec = 0

        use_date = exp_date
        if not exp_date:
            exp_date = opt_dates.option_expiration(
                 date=exp_date)
            use_date = exp_date.strftime('%Y-%m-%d')

        """
        Debug control flags

        Quickly turn specific fetches off:

        get_news = False
        get_pricing = False
        get_options = False

        """
        if get_pricing:
            log.info(
                '{} getting ticker={} pricing'.format(
                    label,
                    ticker))
            ticker_results.get_quotes()
            if ticker_results.quotes_data:
                pricing_dict = ticker_results.quotes_data

                cur_high = pricing_dict.get(
                    'regularMarketDayHigh',
                    None)
                cur_low = pricing_dict.get(
                    'regularMarketDayLow',
                    None)
                cur_open = pricing_dict.get(
                    'regularMarketOpen',
                    None)
                cur_close = pricing_dict.get(
                    'regularMarketPreviousClose',
                    None)
                cur_volume = pricing_dict.get(
                    'regularMarketVolume',
                    None)
                pricing_dict['high'] = cur_high
                pricing_dict['low'] = cur_low
                pricing_dict['open'] = cur_open
                pricing_dict['close'] = cur_close
                pricing_dict['volume'] = cur_volume
                pricing_dict['date'] = get_last_close_str()
                if 'regularMarketTime' in pricing_dict:
                    pricing_dict['market_time'] = \
                        datetime.datetime.fromtimestamp(
                            pricing_dict['regularMarketTime']).strftime(
                                COMMON_TICK_DATE_FORMAT)
                if 'postMarketTime' in pricing_dict:
                    pricing_dict['post_market_time'] = \
                        datetime.datetime.fromtimestamp(
                            pricing_dict['postMarketTime']).strftime(
                                COMMON_TICK_DATE_FORMAT)

                log.info(
                    '{} ticker={} converting pricing to '
                    'df orient={}'.format(
                        label,
                        ticker,
                        orient))

                try:
                    rec['pricing'] = pricing_dict
                except Exception as f:
                    rec['pricing'] = '{}'
                    log.info(
                        '{} ticker={} failed converting pricing '
                        'data={} to df ex={}'.format(
                            label,
                            ticker,
                            ppj(pricing_dict),
                            f))
                # try/ex

                log.info(
                    '{} ticker={} done converting pricing to '
                    'df orient={}'.format(
                        label,
                        ticker,
                        orient))

            else:
                log.error(
                    '{} ticker={} missing quotes_data={}'.format(
                        label,
                        ticker,
                        ticker_results.quotes_data))
            # end of if ticker_results.quotes_data

            log.info(
                '{} ticker={} close={} vol={}'.format(
                    label,
                    ticker,
                    cur_close,
                    cur_volume))
        else:
            log.info(
                '{} skip - getting ticker={} pricing'.format(
                    label,
                    ticker,
                    get_pricing))
        # if get_pricing

        if get_news:
            log.info(
                '{} getting ticker={} news'.format(
                    label,
                    ticker))
            ticker_results.get_news()
            if ticker_results.news_data:
                news_list = None
                try:
                    news_list = ticker_results.news_data
                    log.info(
                        '{} ticker={} converting news to '
                        'df orient={}'.format(
                            label,
                            ticker,
                            orient))

                    num_news_rec = len(news_list)

                    rec['news'] = news_list
                except Exception as f:
                    rec['news'] = '{}'
                    log.info(
                        '{} ticker={} failed converting news '
                        'data={} to df ex={}'.format(
                            label,
                            ticker,
                            news_list,
                            f))
                # try/ex

                log.info(
                    '{} ticker={} done converting news to '
                    'df orient={}'.format(
                        label,
                        ticker,
                        orient))
            else:
                log.info(
                    '{} ticker={} Yahoo NO news={}'.format(
                        label,
                        ticker,
                        ticker_results.news_data))
            # end of if ticker_results.news_data
        else:
            log.info(
                '{} skip - getting ticker={} news'.format(
                    label,
                    ticker))
        # end if get_news

        if get_options:

            get_all_strikes = True
            if get_all_strikes:
                cur_strike = None
            else:
                if cur_close:
                    cur_strike = int(cur_close)
                if not cur_strike:
                    cur_strike = 287

            log.info(
                '{} ticker={} num_news={} get options close={} '
                'exp_date={} contract={} strike={}'.format(
                    label,
                    ticker,
                    num_news_rec,
                    cur_close,
                    use_date,
                    contract_type,
                    cur_strike))

            options_dict = \
                yahoo_get_pricing.get_options(
                    ticker=ticker,
                    exp_date_str=use_date,
                    contract_type=contract_type,
                    strike=cur_strike)

            rec['options'] = '{}'

            try:
                log.info(
                    '{} ticker={} converting options to '
                    'df orient={}'.format(
                        label,
                        ticker,
                        orient))

                num_option_calls = options_dict.get(
                    'num_calls',
                    None)
                num_option_puts = options_dict.get(
                    'num_puts',
                    None)
                rec['options'] = {
                    'exp_date': options_dict.get(
                        'exp_date',
                        None),
                    'calls': options_dict.get(
                        'calls',
                        None),
                    'puts': options_dict.get(
                        'puts',
                        None),
                    'num_calls': num_option_calls,
                    'num_puts': num_option_puts
                }
                rec['calls'] = rec['options'].get(
                    'calls',
                    EMPTY_DF_STR)
                rec['puts'] = rec['options'].get(
                    'puts',
                    EMPTY_DF_STR)
            except Exception as f:
                rec['options'] = '{}'
                log.info(
                    '{} ticker={} failed converting options '
                    'data={} to df ex={}'.format(
                        label,
                        ticker,
                        options_dict,
                        f))
            # try/ex

            log.info(
                '{} ticker={} done converting options to '
                'df orient={} num_calls={} num_puts={}'.format(
                    label,
                    ticker,
                    orient,
                    num_option_calls,
                    num_option_puts))

        else:
            log.info(
                '{} skip - getting ticker={} options'.format(
                    label,
                    ticker))
        # end of if get_options

        log.info(
            '{} yahoo pricing for ticker={} close={} '
            'num_calls={} num_puts={} news={}'.format(
                label,
                ticker,
                cur_close,
                num_option_calls,
                num_option_puts,
                num_news_rec))

        fields_to_upload = [
            'pricing',
            'options',
            'calls',
            'puts',
            'news'
        ]

        for field_name in fields_to_upload:
            upload_and_cache_req = copy.deepcopy(work_dict)
            upload_and_cache_req['celery_disabled'] = True
            upload_and_cache_req['data'] = rec[field_name]
            if not upload_and_cache_req['data']:
                upload_and_cache_req['data'] = '{}'

            if 'redis_key' in work_dict:
                upload_and_cache_req['redis_key'] = '{}_{}'.format(
                    work_dict.get(
                        'redis_key',
                        '{}_{}'.format(
                            ticker,
                            field_name)),
                    field_name)
            if 's3_key' in work_dict:
                upload_and_cache_req['s3_key'] = '{}_{}'.format(
                    work_dict.get(
                        's3_key',
                        '{}_{}'.format(
                            ticker,
                            field_name)),
                    field_name)
            try:
                update_res = publisher.run_publish_pricing_update(
                    work_dict=upload_and_cache_req)
                update_status = update_res.get(
                    'status',
                    NOT_SET)
                log.info(
                    '{} publish update status={} data={}'.format(
                        label,
                        get_status(status=update_status),
                        update_res))
            except Exception as f:
                err = (
                    '{} - failed to upload YAHOO data={} to '
                    'to s3_key={} and redis_key={}'.format(
                        label,
                        upload_and_cache_req,
                        upload_and_cache_req['s3_key'],
                        upload_and_cache_req['redis_key']))
                log.error(err)
            # end of try/ex to upload and cache
            if not rec[field_name]:
                log.debug(
                    '{} - ticker={} no data from YAHOO for '
                    'field_name={}'.format(
                        label,
                        ticker,
                        field_name))
        # end of for all fields

        res = build_result.build_result(
            status=SUCCESS,
            err=None,
            rec=rec)
    except Exception as e:
        res = build_result.build_result(
            status=ERR,
            err=(
                'failed - get_data_from_yahoo '
                'dict={} with ex={}').format(
                    work_dict,
                    e),
            rec=rec)
        log.error(
            '{} - {}'.format(
                label,
                res['err']))
    # end of try/ex

    log.info(
        'task - get_data_from_yahoo done - '
        '{} - status={}'.format(
            label,
            get_status(res['status'])))

    return res
# end of get_data_from_yahoo



"""

Utilities for getting pricing data from yahoo finance for:

- pricing
- news
- options

Internal version of:
https://github.com/neberej/pinance/master/pinance/engine/yfinance2.py
"""


import datetime
import json
import random
import urllib.request
import urllib.error
import urllib.parse
import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def create_url(
        ticker,
        exp_date_str):
    """create_url

    :param ticker: ticker to look up
    :param exp_date_str: expiration
    """
    srv = random.randrange(1, 3, 1)
    if exp_date_str:
        return (
            'https://query{}.finance.yahoo.com/v7/'
            'finance/options/{}?&date={}').format(
                srv,
                ticker,
                exp_date_str)
    else:
        return (
            'https://query{}.finance.yahoo.com/v7/'
            'finance/options/{}').format(
                srv,
                ticker)
# end of create_url


# Convert date/time to unix time for options
def totimestamp(
        inputdate,
        epoch=datetime.datetime(1970, 1, 1)):
    """totimestamp

    :param 1970:
    :param 1:
    :param 1:
    """
    dt = datetime.datetime.strptime(inputdate, '%Y-%m-%d')
    td = dt - epoch
    timestamp = (
        td.microseconds + (
            td.seconds + td.days * 24 * 3600
        ) * 10**6) / 1e6  # td.total_seconds()
    return int(timestamp)
# end of totimestamp


# Make request to yahoo finance
def make_request(
        ticker,
        exp_date_str=None):
    """make_request

    :param ticker: ticker to use
    :param exp_date_str: contract expiration date format ``YYYY-MM-DD``
    """
    if exp_date_str:
        use_exp_date = totimestamp(exp_date_str)
        url = create_url(ticker, use_exp_date)
    else:
        url = create_url(ticker, None)

    try:
        response = json.loads(
            urllib.request.urlopen(url).read().decode('utf-8'))
    except urllib.error.URLError as e:
        if hasattr(e, 'reason'):
            return []
        elif hasattr(e, 'code'):
            return []
    return response
# end of make_request


def extract_options_data(
        response,
        contract_type,
        strike=None):
    """extract_options_data

    :param response: previous response data
    :param contract_type: ``C`` for calls or ``P`` for puts
    :param strike: strike price
    """
    if strike:
        log.debug(
            'getting contract={} strike={}'.format(
                contract_type,
                strike))
        if contract_type == 'C':
            calls = response['optionChain']['result'][0]['options'][0]['calls']
            for call in calls:
                if call['strike'] == round(strike, 1):
                    return [
                        call
                    ]

        elif contract_type == 'P':
            puts = response['optionChain']['result'][0]['options'][0]['puts']
            for put in puts:
                if put['strike'] == round(strike, 1):
                    return [
                        put
                    ]
    else:
        log.debug(
            'getting all chains')
    # end of if strike

    return response['optionChain']['result'][0]['options']
# end of extract_options_data


def get_quotes(
        ticker):
    """get_quotes

    :param ticker: ticker to get pricing data
    """
    response = make_request(
        ticker=ticker,
        exp_date_str=None)
    try:
        quotes_data = response['optionChain']['result'][0]['quote']
        return quotes_data
    except Exception as e:
        log.error(
            'failed get_quotes(ticker={}) with ex={}'.format(
                ticker,
                e))
        return []
# end of get_quotes


def get_options(
        ticker,
        contract_type,
        exp_date_str,
        strike=None):
    """get_options

    :param ticker: ticker to lookup
    :param exp_date_str: ``YYYY-MM-DD`` expiration date format
    :param strike: optional strike price, ``None`` returns
                   all option chains
    :param contract_type: ``C`` calls or ``P`` for puts, if
                          ``strike=None`` then the ``contract_type``
                          is ignored
    """
    log.info(
        'get_options ticker={} '
        'contract={} exp_date={} strike={}'.format(
            ticker,
            contract_type,
            exp_date_str,
            strike))

    response = make_request(
        ticker=ticker,
        exp_date_str=exp_date_str)
    try:
        options_data = extract_options_data(
            response=response,
            contract_type=contract_type,
            strike=strike)
        options_dict = {
            'date': ae_utils.get_last_close_str(),
            'exp_date': None,
            'num_calls': None,
            'num_puts': None,
            'calls': None,
            'puts': None
        }
        if 'expirationDate' in options_data[0]:
            epoch_exp = options_data[0]['expirationDate']
            options_dict['exp_date'] = \
                datetime.datetime.fromtimestamp(
                    epoch_exp).strftime(
                        ae_consts.COMMON_TICK_DATE_FORMAT)
        calls_df = pd.DataFrame(
            options_data[0]['calls'])
        options_dict['num_calls'] = len(
            options_data[0]['calls'])
        options_dict['calls'] = calls_df.to_json(
            orient='records')
        puts_df = pd.DataFrame(
            options_data[0]['puts'])
        options_dict['num_puts'] = len(
            options_data[0]['puts'])
        options_dict['puts'] = puts_df.to_json(
            orient='records')

        return options_dict
    except Exception as e:
        log.error(
            'failed get_options('
            'ticker={}, '
            'contract_type={}, '
            'exp_date_str={}, '
            'strike={}) with ex={}'.format(
                ticker,
                contract_type,
                exp_date_str,
                strike,
                e))
        return []
# end of get_options



"""
FinViz constants and static values
"""

FETCH_SCREENER_TICKERS = 1200

DATAFEED_SCREENER_TICKERS = 1300

DEFAULT_FINVIZ_COLUMNS = [
    'ticker_id',
    'ticker',
    'company',
    'sector',
    'industry',
    'country',
    'market_cap',
    'pe',
    'price',
    'change',
    'volume'
]


def get_ft_str_finviz(
        ft_type):
    """get_ft_str_finviz

    :param ft_type: enum fetch type value to return
                    as a string
    """
    if ft_type == FETCH_SCREENER_TICKERS:
        return 'fv_screener'
    else:
        return 'unsupported ft_type={}'.format(
            ft_type)
    # end of if/else
# end of get_ft_str_finviz


def get_datafeed_str_finviz(
        df_type):
    """get_ft_str_finviz

    :param df_type: enum fetch type value to return
                    as a string
    """
    if df_type == DATAFEED_SCREENER_TICKERS:
        return 'fv_screener'
    else:
        return 'unsupported df_type={}'.format(
            df_type)
# end of get_datafeed_str_finviz



"""
Base Indicator Class for deriving your own indicators
to use within an ``analysis_engine.indicators.in
dicator_processor.IndicatorProcessor``
"""

import uuid
import pandas as pd
import logging
import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils


class BaseIndicator:
    """BaseIndicator"""

    def __init__(
            self,
            config_dict,
            path_to_module=None,
            name=None,
            verbose=False):
        """__init__

        Base class for building your own indicators to work
        within an ``IndicatorProcessor``.

        Please derive the ``self.process()`` method as needed

        .. tip:: any keys passed in with ``config_dict`` will
            become class member variables that can be accessed
            and used as normal member variables within the
            derived Indicator class

        :param config_dict: dictionary for this indicator
        :param name: name of the indicator
        :param path_to_module: work in progress -
            this will allow loading indicators from
            outside the repo like the derived algorithm
            classes
        :param verbose: optional - bool for toggling more logs
        """
        self.name = name
        self.log = log_utils.build_colorized_logger(
            name=name)

        self.config = config_dict
        self.path_to_module = path_to_module
        self.verbose = verbose

        if not self.config:
            raise Exception(
                'please provide a config_dict for loading '
                'the buy and sell rules for this indicator')

        if not self.verbose:
            self.verbose = self.config.get(
                'verbose',
                False)

        if not self.name:
            self.name = 'ind_{}'.format(
                str(uuid.uuid4()).replace('-', ''))

        self.starter_dict = None
        self.previous_df = self.config.get(
            'previous_df',
            None)
        self.name_of_df = self.config.get(
            'uses_data',
            None)
        self.uses_data = self.name_of_df
        self.report = self.config.get(
            'report',
            {})
        self.ind_id = self.report.get(
            'id',
            self.name)
        self.metrics = self.report.get(
            'metrics',
            {})
        self.ind_type = self.metrics.get(
            'type',
            ae_consts.INDICATOR_TYPE_UNKNOWN)
        self.ind_category = self.metrics.get(
            'category',
            ae_consts.INDICATOR_CATEGORY_UNKNOWN)
        self.ind_uses_data = self.metrics.get(
            'ind_uses_data',
            ae_consts.INDICATOR_USES_DATA_ANY)
        self.dataset_df_str = self.config.get(
            'dataset_df',
            None)

        self.report_key_prefix = self.report.get(
            'report_key_prefix',
            self.name)

        # this should be mostly numeric values
        # to allow converting to an AI-ready dataset
        # once the algorithm finishes
        self.report_dict = {
            'type': self.ind_type,
            'category': self.ind_category,
            'uses_data': self.ind_uses_data
        }

        self.report_ignore_keys = self.config.get(
            'report_ignore_keys',
            ae_consts.INDICATOR_IGNORED_CONIGURABLE_KEYS)
        self.use_df = pd.DataFrame(
            ae_consts.EMPTY_DF_LIST)
        self.configurables = self.config
        self.ind_confs = []
        self.convert_config_keys_to_members()
    # end of __init__

    def get_config(
            self):
        """get_config"""
        pruned_config = {}

        # remove the obj
        remove_keys = [
            'obj'
        ]
        for k in self.config:
            if k not in remove_keys:
                pruned_config[k] = self.config[k]
        return pruned_config
    # end of get_config

    def convert_config_keys_to_members(
            self):
        """convert_config_keys_to_members

        This converts any key in the config to
        a member variable that can be used with the
        your derived indicators like: ``self.<KEY_IN_CONFIG>``
        """
        for k in self.config:
            if k not in self.report_ignore_keys:
                self.__dict__[k] = self.config[k]
    # end of convert_config_keys_to_members

    def build_configurable_node(
            self,
            name,
            conf_type,
            current_value=None,
            default_value=None,
            max_value=None,
            min_value=None,
            is_output_only=False,
            inc_interval=None,
            notes=None,
            **kwargs):
        """build_configurable_node

        Helper for building a single configurable type
        node for programmatically creating algo configs

        :param name: name of the member configurable
        :param conf_type: string - configurable type
        :param current_value: optional - current value
        :param default_value: optional - default value
        :param max_value: optional - maximum value
        :param min_value: optional - minimum value
        :param is_output_only: optional - bool for setting
            the input parameter as an output-only value
            (default is ``False``)
        :param inc_interval: optional - float value
            for controlling how the tests should increment
            while walking between the ``min_value`` and the
            ``max_value``
        :param notes: optional - string notes
        :param kwargs: optional - derived keyword args dictionary
        """
        node = {
            'name': name,
            'type': conf_type,
            'value': current_value,
            'default': default_value,
            'max': max_value,
            'min': min_value,
            'is_output_only': is_output_only,
            'inc_interval': inc_interval,
            'notes': notes
        }
        for k in kwargs:
            node[k] = kwargs[k]
        return node
    # end of build_configurable_node

    def build_base_configurables(
            self,
            ind_type='momentum',
            category='technical',
            uses_data='minute',
            version=1):
        """build_base_configurables

        :param ind_type: string indicator type
        :param category: string indicator category
        :param uses_data: string indicator
            usess this type of data
        :param version: integer for building
            configurables for the testing
            generation version
        """
        self.ind_confs = []
        self.ind_confs.append(self.build_configurable_node(
            name='category',
            conf_type='str',
            default_value=category,
            is_output_only=True))
        self.ind_confs.append(self.build_configurable_node(
            name='type',
            conf_type='str',
            default_value=ind_type,
            is_output_only=True))
        self.ind_confs.append(self.build_configurable_node(
            name='uses_data',
            conf_type='str',
            default_value=self.config.get(
                'uses_data',
                uses_data),
            is_output_only=True))

        if version == 1:
            self.ind_confs.append(self.build_configurable_node(
                name='is_buy',
                conf_type='int',
                is_output_only=True))
            self.ind_confs.append(self.build_configurable_node(
                name='is_sell',
                conf_type='int',
                is_output_only=True))
    # end of build_base_configurables

    def get_configurables(
            self,
            **kwargs):
        """get_configurables

        **Derive this in your indicators**

        This is used as a helper for setting up algorithm
        configs for this indicator and to programmatically set
        the values based off the domain rules

        :param kwargs: optional keyword args
        """
        self.ind_confs = []

        self.lg(
            'configurables={} for class={}'.format(
                ae_consts.ppj(self.ind_confs),
                self.__class__.__name__))

        return self.ind_confs
    # end of get_configurables

    def set_configurables(
            self,
            config_dict):
        """set_configurables

        :param config_dict: indicator config dictionary
        """
        self.configurables = config_dict
        return self.configurables
    # end of set_configurables

    def lg(
            self,
            msg,
            level=logging.INFO):
        """lg

        Log only if the indicator has ``self.verbose``
        set to ``True`` or if the ``level == logging.CRITICAL`` or
        ``level = logging.ERROR`` otherwise no logs

        :param msg: string message to log
        :param level: set the logging level
            (default is ``logging.INFO``)
        """
        if self.verbose:
            if level == logging.INFO:
                self.log.info(msg)
                return
            elif level == logging.ERROR:
                self.log.error(msg)
                return
            elif level == logging.DEBUG:
                self.log.debug(msg)
                return
            elif level == logging.WARN:
                self.log.warn(msg)
                return
            elif level == logging.CRITICAL:
                self.log.critical(msg)
                return
        else:
            if level == logging.ERROR:
                self.log.error(msg)
                return
            elif level == logging.CRITICAL:
                self.log.critical(msg)
                return
    # end lg

    def get_report_prefix(
            self):
        """get_report_prefix"""
        return self.report_key_prefix
    # end of get_report_prefix

    def build_report_key(
            self,
            key,
            prefix_key,
            key_type,
            cur_report_dict):
        """build_report_key

        :param prefix_key:
        """
        report_key = (
            f'{prefix_key}_{key}')
        if report_key in cur_report_dict:
            report_key = (
                f'{report_key}_'
                f'{str(uuid.uuid4()).replace("-", "")}')
        # end of building a key to prevent stomping data

        return report_key
    # end of build_report_key

    def get_report(
            self,
            verbose=False):
        """get_report

        Get the indicator's current output node
        that is used for the trading performance report
        generated at the end of the algorithm

        .. note:: the report dict should mostly be numeric
            types to enable AI predictions after removing
            non-numeric columns

        :param verbose: optional - boolean for toggling
            to show the report
        """
        cur_report_dict = {}

        # allow derived indicators to build their own report prefix
        report_prefix_key_name = self.get_report_prefix()

        for key in self.report_dict:

            is_valid = True
            if key in self.report_ignore_keys:
                is_valid = False

            if is_valid:
                report_key = self.build_report_key(
                    key,
                    prefix_key=report_prefix_key_name,
                    key_type='report',
                    cur_report_dict=cur_report_dict)
                cur_report_dict[report_key] = self.report_dict[key]
        # for all keys to output into the report

        buy_value = None
        sell_value = None

        for key in self.configurables:

            is_valid = True
            if key in self.report_ignore_keys:
                is_valid = False
            elif key not in self.__dict__:
                is_valid = False

            if is_valid:
                report_key = self.build_report_key(
                    key,
                    prefix_key=report_prefix_key_name,
                    key_type='conf',
                    cur_report_dict=cur_report_dict)

                use_value = None
                if key == 'is_buy':
                    buy_value = self.__dict__[key]
                    if buy_value:
                        use_value = \
                            ae_consts.INDICATOR_ACTIONS[buy_value]
                    else:
                        use_value = \
                            ae_consts.INT_INDICATOR_NOT_PROCESSED
                elif key == 'is_sell':
                    sell_value = self.__dict__[key]
                    if sell_value:
                        use_value = \
                            ae_consts.INDICATOR_ACTIONS[sell_value]
                    else:
                        use_value = \
                            ae_consts.INT_INDICATOR_NOT_PROCESSED
                else:
                    use_value = self.__dict__[key]
                # end of deciding value

                cur_report_dict[report_key] = use_value
            # if valid

        # end of all configurables for this indicator

        if verbose or self.verbose:
            self.lg(
                'indicator={} '
                'report={} '
                'buy={} sell={}'.format(
                    self.name,
                    ae_consts.ppj(cur_report_dict),
                    buy_value,
                    sell_value))

        return cur_report_dict
    # end of get_report

    def get_path_to_module(
            self):
        """get_path_to_module"""
        return self.path_to_module
    # end of get_path_to_module

    def get_name(
            self):
        """get_name"""
        return self.name
    # end of get_name

    def get_dataset_by_name(
            self,
            dataset,
            dataset_name):
        """get_dataset_by_name

        Method for getting just a dataset
        by the dataset_name`` inside the cached
        ``dataset['data']`` dictionary of ``pd.Dataframe(s)``

        :param dataset: cached dataset value
            that holds the dictionaries: ``dataset['data']``
        :param dataset_name: optional - name of the
            supported ``pd.DataFrame`` that is in the
            cached ``dataset['data']`` dictionary
            of dataframes
        """
        return dataset['data'].get(
            dataset_name,
            pd.DataFrame(ae_consts.EMPTY_DF_LIST))
    # end of get_dataset_by_name

    def get_subscribed_dataset(
            self,
            dataset,
            dataset_name=None):
        """get_subscribed_dataset

        Method for getting just the subscribed dataset
        else use the ``dataset_name`` argument dataset

        :param dataset: cached dataset value
            that holds the dictionaries: ``dataset['data']``
        :param dataset_name: optional - name of the
            supported ``pd.DataFrame`` that is in the
            cached ``dataset['data']`` dictionary
            of dataframes
        """
        ret_df = None
        if dataset_name:
            ret_df = dataset['data'].get(
                dataset_name,
                pd.DataFrame(ae_consts.EMPTY_DF_LIST))
        else:
            ret_df = dataset['data'].get(
                self.name_of_df,
                pd.DataFrame(ae_consts.EMPTY_DF_LIST))

        if hasattr(ret_df, 'index'):
            return ae_consts.SUCCESS, ret_df
        else:
            return ae_consts.EMPTY, ret_df
    # end of get_subscribed_dataset

    def reset_internals(
            self,
            **kwargs):
        """reset_internals

        Support a cleanup action before indicators
        run between datasets. Derived classes can
        implement custom cleanup actions that need
        to run before each ``IndicatorProcessor.process()``
        call is run on the next cached dataset

        :param kwargs: keyword args dictionary
        """
        return ae_consts.SUCCESS
    # end of reset_internals

    def handle_subscribed_dataset(
            self,
            algo_id,
            ticker,
            dataset):
        """handle_subscribed_dataset

        Filter the algorithm's ``dataset`` to just the
        dataset the indicator is set up to use as defined by
        the member variable:

        - ``self.name_of_df`` - string value like ``daily``, ``minute``

        :param algo_id: string - algo identifier label for debugging datasets
            during specific dates
        :param ticker: string - ticker
        :param dataset: dictionary of ``pd.DataFrame(s)`` to process
        """

        # certain datasets like minutes or options may
        # want to refer to the previous dataset
        self.previous_df = dataset

        # call derived class's process()
        self.process(
            algo_id=algo_id,
            ticker=ticker,
            dataset=dataset)
    # end of handle_subscribed_dataset

    def process(
            self,
            algo_id,
            ticker,
            dataset):
        """process

        Derive custom indicator processing to determine buy and sell
        conditions before placing orders. Just implement your own
        ``process`` method.

        Please refer to the TA Lib guides for details on building indicators:

        - Overlap Studies
          https://mrjbq7.github.io/ta-lib/func_groups/overlap_studies.html
        - Momentum Indicators
          https://mrjbq7.github.io/ta-lib/func_groups/momentum_indicators.html
        - Volume Indicators
          https://mrjbq7.github.io/ta-lib/func_groups/volume_indicators.html
        - Volatility Indicators
          https://mrjbq7.github.io/ta-lib/func_groups/volatility_indicators.html
        - Price Transform
          https://mrjbq7.github.io/ta-lib/func_groups/price_transform.html
        - Cycle Indicators
          https://mrjbq7.github.io/ta-lib/func_groups/cycle_indicators.html
        - Pattern Recognition
          https://mrjbq7.github.io/ta-lib/func_groups/pattern_recognition.html
        - Statistic Functions
          https://mrjbq7.github.io/ta-lib/func_groups/statistic_functions.html
        - Math Transform
          https://mrjbq7.github.io/ta-lib/func_groups/math_transform.html
        - Math Operators
          https://mrjbq7.github.io/ta-lib/func_groups/math_operators.html

        :param algo_id: string - algo identifier label for debugging datasets
            during specific dates
        :param ticker: string - ticker
        :param dataset: dictionary of ``pd.DataFrame(s)`` to process
        """
        self.lg(
            '{} BASE_IND process - start'.format(
                self.name))
        self.lg(
            '{} BASE_IND process - end'.format(
                self.name))
    # end of process

# end of BaseIndicator



"""
Get a Celery Application Helper
"""

import os
import celery
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def get_celery_app(
        name=os.getenv(
            'APP_NAME',
            'worker'),
        auth_url=os.getenv(
            'WORKER_BROKER_URL',
            'redis://localhost:6379/11'),
        backend_url=os.getenv(
            'WORKER_BACKEND_URL',
            'redis://localhost:6379/12'),
        include_tasks=[],
        ssl_options=None,
        transport_options=None,
        path_to_config_module=os.getenv(
            'WORKER_CELERY_CONFIG_MODULE',
            'analysis_engine.work_tasks.celery_config'),
        worker_log_format=os.getenv(
            'WORKER_LOG_FORMAT',
            '%(asctime)s: %(levelname)s %(message)s'),
        **kwargs):
    """get_celery_app

    Build a Celery app with support for environment variables
    to set endpoints locations.

    - export WORKER_BROKER_URL=redis://localhost:6379/11
    - export WORKER_BACKEND_URL=redis://localhost:6379/12
    - export WORKER_CELERY_CONFIG_MODULE=analysis_engine.work_tasks.cel
      ery_config

    .. note:: Jupyter notebooks need to use the
        ``WORKER_CELERY_CONFIG_MODULE=analysis_engine.work_tasks.celery
        service_config`` value which uses resolvable hostnames with
        docker compose:

        - export WORKER_BROKER_URL=redis://redis:6379/11
        - export WORKER_BACKEND_URL=redis://redis:6379/12

    :param name: name for this app
    :param auth_url: Celery broker address
        (default is ``redis://localhost:6379/11``
        or ``analysis_engine.consts.WORKER_BROKER_URL``
        environment variable)
        this is required for distributing algorithms
    :param backend_url: Celery backend address
        (default is ``redis://localhost:6379/12``
        or ``analysis_engine.consts.WORKER_BACKEND_URL``
        environment variable)
        this is required for distributing algorithms
    :param include_tasks: list of modules containing tasks to add
    :param ssl_options: security options dictionary
        (default is ``analysis_engine.consts.SSL_OPTIONS``)
    :param trasport_options: transport options dictionary
        (default is ``analysis_engine.consts.TRANSPORT_OPTIONS``)
    :param path_to_config_module: config module for advanced
        Celery worker connectivity requirements
        (default is ``analysis_engine.work_tasks.celery_config``
        or ``analysis_engine.consts.WORKER_CELERY_CONFIG_MODULE``)
    :param worker_log_format: format for logs
    """

    if len(include_tasks) == 0:
        log.error(
            'creating celery app={} MISSING tasks={}'.format(
                name,
                include_tasks))
    else:
        log.info(
            'creating celery app={} tasks={}'.format(
                name,
                include_tasks))

    # get the Celery application
    app = celery.Celery(
        name,
        broker_url=auth_url,
        result_backend=backend_url,
        include=include_tasks)

    app.config_from_object(
        path_to_config_module,
        namespace='CELERY')

    app.conf.update(kwargs)

    if transport_options:
        log.info(
            'loading transport_options={}'.format(
                transport_options))
        app.conf.update(**transport_options)
    # custom tranport options

    if ssl_options:
        log.info(
            'loading ssl_options={}'.format(
                ssl_options))
        app.conf.update(**ssl_options)
    # custom ssl options

    if len(include_tasks) > 0:
        app.autodiscover_tasks(include_tasks)

    return app
# end of get_celery_app



"""
Build a scaler normalized
``pandas.DataFrame`` from an existing ``pandas.DataFrame``
"""

import sklearn.preprocessing as preproc
import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(
    name=__name__)


def build_scaler_dataset_from_df(
        df,
        min_feature=-1,
        max_feature=1):
    """build_scaler_dataset_from_df

    Helper for building scaler datasets from an existing ``pandas.DataFrame``

    returns a dictionary:

    .. code-block:: python

        return {
            'status': status,   # NOT_RUN | SUCCESS | ERR
            'scaler': scaler,   # MinMaxScaler
            'df': df  # scaled df from df arg
        }

    :param df: ``pandas.DataFrame`` to convert to scalers
    :param min_feature: min feature range for scaler normalization
        with default ``-1``
    :param max_feature: max feature range for scaler normalization
        with default ``1``
    """

    status = ae_consts.NOT_RUN
    scaler = None
    output_df = None

    res = {
        'status': status,
        'scaler': scaler,
        'df': output_df
    }

    try:
        log.debug(
            f'building scaler range=[{min_feature}, {max_feature}]')
        scaler = preproc.MinMaxScaler(
            feature_range=(
                min_feature,
                max_feature))

        log.info(
            f'scaler.fit_transform(df) rows={len(df.index)}')
        output_df = scaler.fit_transform(
            df.values)
        status = ae_consts.SUCCESS
    except Exception as e:
        log.error(
            f'failed build_scaler_dataset_from_df '
            f'with ex={e} '
            f'range=[{min_feature}, {max_feature}]')
        status = ae_consts.ERR
    # end of try/ex

    res = {
        'status': status,
        'scaler': scaler,
        'df': output_df
    }
    return res
# end of build_scaler_dataset_from_df



"""
Perform dataset scrubbing actions
and return the scrubbed dataset as a ready-to-go
data feed. This is an approach for normalizing
an internal data feed.

Supported environment variables:

::

    # verbose logging in this module
    # note this can take longer to transform
    # DataFrames and is not recommended for
    # production:
    export DEBUG_FETCH=1

Ingress Scrubbing supports converting an incoming
dataset (from IEX) and converts it to one
of the following data feed and returned as a
``pandas DataFrame``:

.. code-block:: python

    DATAFEED_DAILY = 900
    DATAFEED_MINUTE = 901
    DATAFEED_QUOTE = 902
    DATAFEED_STATS = 903
    DATAFEED_PEERS = 904
    DATAFEED_NEWS = 905
    DATAFEED_FINANCIALS = 906
    DATAFEED_EARNINGS = 907
    DATAFEED_DIVIDENDS = 908
    DATAFEED_COMPANY = 909
    DATAFEED_PRICING_YAHOO = 1100
    DATAFEED_OPTIONS_YAHOO = 1101
    DATAFEED_NEWS_YAHOO = 1102

"""

import datetime
import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.iex.consts as iex_consts
import analysis_engine.yahoo.consts as yahoo_consts
import analysis_engine.td.consts as td_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def debug_msg(
        label,
        datafeed_type,
        msg_format,
        date_str,
        df):
    """debug_msg

    Debug helper for debugging scrubbing handlers

    :param label: log label
    :param datafeed_type: fetch type
    :param msg_format: message to include
    :param date_str: date string
    :param df: ``pandas DataFrame`` or ``None``
    """

    msg = msg_format.format('_', date_str)

    dft_msg = ''
    if (
            datafeed_type == yahoo_consts.DATAFEED_PRICING_YAHOO or
            datafeed_type == yahoo_consts.DATAFEED_OPTIONS_YAHOO or
            datafeed_type == yahoo_consts.DATAFEED_NEWS_YAHOO):
        dft_msg = yahoo_consts.get_datafeed_str_yahoo(
            df_type=datafeed_type)
    elif (
            datafeed_type == td_consts.DATAFEED_TD_CALLS or
            datafeed_type == td_consts.DATAFEED_TD_PUTS):
        dft_msg = td_consts.get_datafeed_str_td(
            df_type=datafeed_type)
    else:
        dft_msg = iex_consts.get_datafeed_str(
            df_type=datafeed_type)

    if ae_consts.ev('DEBUG_FETCH', '0') == '1':
        if 'START' in msg:
            log.info(
                f'{label} - {dft_msg} '
                f'-------------------------'
                f'------------------------------------')
        msg = msg_format.format(
            df,
            date_str),
        if hasattr(df, 'empty'):
            log.info(
                f'{label} - {dft_msg} - {msg} found df={df} '
                f'columns={df.columns.values}')
        else:
            log.info(
                f'{label} - {dft_msg} - {msg} found df={df}')

        if 'END' in msg:
            log.info(
                f'{label} - {dft_msg} '
                f'-------------------------'
                f'------------------------------------')
    else:
        log.info(
            f'{label} - {dft_msg} - {msg}')
    # end of debug pre-scrub logging

# end of debug_msg


def build_dates_from_df_col(
        df,
        use_date_str,
        src_col='minute',
        src_date_format=ae_consts.COMMON_TICK_DATE_FORMAT,
        output_date_format=ae_consts.COMMON_TICK_DATE_FORMAT):
    """build_dates_from_df_col

    Converts a string date column series in a ``pandas.DataFrame``
    to a well-formed date string list.

    :param src_col: source column name
    :param use_date_str: date string for today
    :param src_date_format: format of the string in the
                            ```df[src_col]`` columne
    :param output_date_format: write the new date strings
                               in this format.
    :param df: source ``pandas.DataFrame``
    """
    new_dates = []
    for idx, i in enumerate(df[src_col]):
        org_new_str = ''
        if ':' not in i:
            split_arr = i.split(' ')
            org_new_str = (
                f'{use_date_str} '
                f'{split_arr[0]}'
                f':00:00 '
                f'{split_arr[1]}')
        else:
            org_new_str = (
                f'{use_date_str} '
                f'{i}:00')
        new_date_val = datetime.datetime.strptime(
            org_new_str,
            src_date_format)
        new_str = new_date_val.strftime(
            output_date_format)
        new_dates.append(new_str)
    # for all rows

    return new_dates
# end of build_dates_from_df_col


def ingress_scrub_dataset(
        label,
        datafeed_type,
        df,
        date_str=None,
        msg_format=None,
        scrub_mode='sort-by-date',
        ds_id='no-id'):
    """ingress_scrub_dataset

    Scrub a ``pandas.DataFrame`` from an Ingress pricing service
    and return the resulting ``pandas.DataFrame``

    :param label: log label
    :param datafeed_type: ``analysis_engine.iex.consts.DATAFEED_*`` type
        or ``analysis_engine.yahoo.consts.DATAFEED_*```
        type
        .. code-block:: python

            DATAFEED_DAILY = 900
            DATAFEED_MINUTE = 901
            DATAFEED_QUOTE = 902
            DATAFEED_STATS = 903
            DATAFEED_PEERS = 904
            DATAFEED_NEWS = 905
            DATAFEED_FINANCIALS = 906
            DATAFEED_EARNINGS = 907
            DATAFEED_DIVIDENDS = 908
            DATAFEED_COMPANY = 909
            DATAFEED_PRICING_YAHOO = 1100
            DATAFEED_OPTIONS_YAHOO = 1101
            DATAFEED_NEWS_YAHOO = 1102

    :param df: ``pandas DataFrame``
    :param date_str: date string for simulating historical dates
                     or ``datetime.datetime.now()`` if not
                     set
    :param msg_format: msg format for a ``string.format()``
    :param scrub_mode: mode to scrub this dataset
    :param ds_id: dataset identifier
    """

    if not hasattr(df, 'empty'):
        log.debug(
            f'{label} - {datafeed_type} no dataset_id={ds_id}')
        return None

    out_df = df

    daily_date_format = '%I:%M %p'
    minute_date_format = '%I:%M %p'

    use_date_str = date_str
    last_close_date = ae_utils.last_close()
    today_str = last_close_date.strftime('%Y-%m-%d')

    year_str = today_str.split('-')[0]
    if not use_date_str:
        use_date_str = today_str

    daily_date_format = ae_consts.IEX_DAILY_DATE_FORMAT
    minute_date_format = ae_consts.IEX_MINUTE_DATE_FORMAT

    """
    use_msg_format = msg_format
    if not msg_format:
        use_msg_format = 'df={} date_str={}'

    debug_msg(
        label=label,
        datafeed_type=datafeed_type,
        msg_format=f'START - {use_msg_format}',
        date_str=use_date_str,
        df=df)
    """

    try:
        if scrub_mode == 'sort-by-date':
            if datafeed_type == iex_consts.DATAFEED_DAILY:
                new_dates = []
                if 'label' in df:
                    for idx, i in enumerate(out_df['label']):
                        split_arr = i.split(' ')
                        new_str = ''
                        if ',' not in i:
                            # Oct 3
                            new_str = (
                                f'{year_str}-{split_arr[0]}-{split_arr[1]}')
                        else:
                            # Aug 29, 18
                            new_str = (
                                f'20{split_arr[2]}-{split_arr[0]}-'
                                f'{split_arr[1].replace(",", "")}')
                        new_dates.append(new_str)
                    # end for all rows
                    out_df['date'] = pd.to_datetime(
                        new_dates,
                        format=daily_date_format)
                # end if label is in df
            elif datafeed_type == iex_consts.DATAFEED_MINUTE:
                new_dates = []
                if 'label' in df:
                    new_dates = build_dates_from_df_col(
                        src_col='label',
                        src_date_format=minute_date_format,
                        use_date_str=use_date_str,
                        df=out_df)
                    out_df['date'] = pd.to_datetime(
                        new_dates,
                        format='%Y-%m-%d %H:%M:%S')
                # end if label is in df
            elif datafeed_type == iex_consts.DATAFEED_QUOTE:
                columns_list = out_df.columns.values
                if 'latestTime' in columns_list:
                    out_df['date'] = pd.to_datetime(
                        out_df['latestTime'],
                        format=ae_consts.IEX_QUOTE_DATE_FORMAT)
                if 'latestUpdate' in columns_list:
                    out_df['latest_update'] = pd.to_datetime(
                        out_df['latestUpdate'],
                        unit='ns')
                if 'extendedPriceTime' in columns_list:
                    out_df['extended_price_time'] = pd.to_datetime(
                        out_df['extendedPriceTime'],
                        unit='ns')
                if 'iexLastUpdated' in columns_list:
                    out_df['iex_last_update'] = pd.to_datetime(
                        out_df['iexLastUpdated'],
                        unit='ns')
                if 'openTime' in columns_list:
                    out_df['open_time'] = pd.to_datetime(
                        out_df['openTime'],
                        unit='ns')
                if 'closeTime' in columns_list:
                    out_df['close_time'] = pd.to_datetime(
                        out_df['closeTime'],
                        unit='ns')
                # end if label is in df
            elif datafeed_type == iex_consts.DATAFEED_STATS:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            elif datafeed_type == iex_consts.DATAFEED_PEERS:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            elif datafeed_type == iex_consts.DATAFEED_NEWS:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            elif datafeed_type == iex_consts.DATAFEED_FINANCIALS:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            elif datafeed_type == iex_consts.DATAFEED_EARNINGS:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            elif datafeed_type == iex_consts.DATAFEED_DIVIDENDS:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            elif datafeed_type == iex_consts.DATAFEED_COMPANY:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            elif datafeed_type == yahoo_consts.DATAFEED_PRICING_YAHOO:
                if 'date' in df:
                    out_df['date'] = pd.to_datetime(
                        df['date'],
                        format=daily_date_format)
            elif datafeed_type == yahoo_consts.DATAFEED_OPTIONS_YAHOO:
                if 'date' in df:
                    out_df['date'] = pd.to_datetime(
                        df['date'],
                        format=daily_date_format)
            elif datafeed_type == yahoo_consts.DATAFEED_NEWS_YAHOO:
                if 'date' in df:
                    out_df['date'] = pd.to_datetime(
                        df['date'],
                        format=daily_date_format)
            elif datafeed_type == td_consts.DATAFEED_TD_CALLS:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            elif datafeed_type == td_consts.DATAFEED_TD_PUTS:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            else:
                log.debug('{label} - {datafeed_type} - no scrub_mode')
            # if/else
        else:
            log.debug(
                f'{label} - {datafeed_type} - '
                f'missing support in ingress_scrub_dataset')
    except Exception as e:
        log.critical(
            f'{label} - {datafeed_type} sort={scrub_mode} - '
            f'failed with ex={e} data={df}')
        out_df = None
    # end of try/ex

    """
    debug_msg(
        label=label,
        datafeed_type=datafeed_type,
        msg_format='END - df={} date_str={}',
        date_str=use_date_str,
        df=out_df)
    """

    return out_df
# end of ingress_scrub_dataset


def extract_scrub_dataset(
        label,
        datafeed_type,
        df,
        date_str=None,
        msg_format=None,
        scrub_mode='sort-by-date',
        ds_id='no-id'):
    """extract_scrub_dataset

    Scrub a cached ``pandas.DataFrame`` that was stored
    in Redis and return the resulting ``pandas.DataFrame``

    :param label: log label
    :param datafeed_type: ``analysis_engine.iex.consts.DATAFEED_*`` type
        or ``analysis_engine.yahoo.consts.DATAFEED_*```
        type
        .. code-block:: python

            DATAFEED_DAILY = 900
            DATAFEED_MINUTE = 901
            DATAFEED_QUOTE = 902
            DATAFEED_STATS = 903
            DATAFEED_PEERS = 904
            DATAFEED_NEWS = 905
            DATAFEED_FINANCIALS = 906
            DATAFEED_EARNINGS = 907
            DATAFEED_DIVIDENDS = 908
            DATAFEED_COMPANY = 909
            DATAFEED_PRICING_YAHOO = 1100
            DATAFEED_OPTIONS_YAHOO = 1101
            DATAFEED_NEWS_YAHOO = 1102
    :param df: ``pandas DataFrame``
    :param date_str: date string for simulating historical dates
                     or ``datetime.datetime.now()`` if not
                     set
    :param msg_format: msg format for a ``string.format()``
    :param scrub_mode: mode to scrub this dataset
    :param ds_id: dataset identifier
    """

    if not hasattr(df, 'empty'):
        log.info(
            f'{label} - {datafeed_type} no dataset_id={ds_id}')
        return None

    return df
# end of extract_scrub_dataset



"""
Functions for getting data from IEX using HTTP
"""

try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse as urlparse
import requests
import pandas as pd
import analysis_engine.iex.consts as iex_consts
import analysis_engine.iex.build_auth_url as iex_auth
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def convert_datetime_columns(
        df,
        date_cols=None,
        second_cols=None,
        tcols=None,
        ecols=None):
    """convert_datetime_columns

    Convert the IEX date columns in the ``df`` to ``datetime`` objects

    :param df: ``pandas.DataFrame`` to set columns to
        datetime objects
    :param date_cols: list of columns to convert with a date string format
        formatted: ``YYYY-MM-DD``
    :param second_cols: list of columns to convert with a date string format
        formatted: ``YYYY-MM-DD HH:MM:SS``
    :param tcols: list of columns to convert with a time format
        (this is for millisecond epoch integers)
    :param ecols: list of columns to convert with a time format
        (this is for nanosecond epoch integers)
    """
    date_cols = date_cols or iex_consts.IEX_DATE_FIELDS
    second_cols = second_cols or iex_consts.IEX_SECOND_FIELDS
    tcols = tcols or iex_consts.IEX_TIME_FIELDS
    ecols = ecols or iex_consts.IEX_EPOCH_FIELDS

    for col in date_cols:
        if col in df:
            df[col] = pd.to_datetime(
                df[col],
                format=iex_consts.IEX_DATE_FORMAT,
                errors='coerce')

    for col in second_cols:
        if col in df:
            df[col] = pd.to_datetime(
                df[col],
                format=iex_consts.IEX_TICK_FORMAT,
                errors='coerce')

    for tcol in tcols:
        if tcol in df:
            df[tcol] = pd.to_datetime(
                df[tcol],
                unit='ms',
                errors='coerce')

    for ecol in ecols:
        if ecol in df:
            df[ecol] = pd.to_datetime(
                df[ecol],
                unit='ns',
                errors='coerce')
# end of convert_datetime_columns


def get_from_iex_v1(
        url):
    """get_from_iex_v1

    Get data from the IEX Trading API (v1)
    https//api.iextrading.com/1.0/

    :param url: IEX V1 Resource URL
    """
    url = (
        f'{iex_consts.IEX_URL_BASE_V1}{url}')
    resp = requests.get(
        urlparse(url).geturl(),
        proxies=iex_consts.IEX_PROXIES)
    if resp.status_code == 200:
        return resp.json()
    raise Exception(
        f'Failed to get data from IEX V1 API with '
        f'function=get_from_iex_v1 '
        f'url={url} which sent '
        f'response {resp.status_code} - {resp.text}')
# end of get_from_iex_v1


def get_from_iex_cloud(
        url,
        token=None):
    """get_from_iex_cloud

    Get data from IEX Cloud API (v2)
    https://iexcloud.io

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    """
    url = (
        f'{iex_consts.IEX_URL_BASE}{url}')
    resp = requests.get(
        url,
        proxies=iex_consts.IEX_PROXIES)
    if resp.status_code == requests.codes.OK:
        return resp.json()
    raise Exception(
        f'Failed to get data from IEX Cloud with '
        f'function=get_from_iex_cloud '
        f'url={url} which sent '
        f'response {resp.status_code} - {resp.text}')
# end of get_from_iex_cloud


def handle_get_from_iex(
        url,
        token=None,
        version=None):
    """handle_get_from_iex

    Implementation for getting data from the IEX
    v2 or v1 api depending on if the ``token``
    argument is set:

    - `IEX Cloud (v2) <https://iexcloud.io>`__
    - `IEX Trading API (v1) <https://iextrading.com/developer/docs/>`__

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :version: optional - string version for the IEX Cloud
        (default is ``beta``)
    """
    if token:
        return get_from_iex_cloud(
            url=url,
            token=token)
    return get_from_iex_v1(
        url=url)
# handle_get_from_iex


def get_from_iex(
        url,
        token=None,
        version=None):
    """get_from_iex

    Helper for getting data from an IEX
    publishable API endpoint using a token
    as a query param on the http url.

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    full_url = iex_auth.build_auth_url(
        url=url,
        token=token)
    return handle_get_from_iex(
        url=full_url,
        token=token,
        version=version)
# end of get_from_iex



"""
Fetch API calls for pulling IEX Cloud Data from
a valid IEX account
"""

import pandas as pd
import analysis_engine.utils as ae_utils
import analysis_engine.dataset_scrub_utils as dataset_utils
import analysis_engine.iex.consts as iex_consts
import analysis_engine.iex.helpers_for_iex_api as iex_helpers
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def fetch_daily(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_daily

    Fetch the IEX daily data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#historical-prices

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/chart/1m')

    log.debug(
        f'{label} - daily - url={use_url} '
        f'args={work_dict} '
        f'ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame(resp_json)

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = [
        'uHigh',
        'uLow',
        'uOpen',
        'uClose',
        'uVolume'
    ]

    remove_these = None
    for c in df:
        if c in cols_to_drop:
            if not remove_these:
                remove_these = []
            remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    df.set_index(
        [
            'date'
        ]).sort_values(by=['date'], ascending=True)

    return df
# end of fetch_daily


def fetch_minute(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_minute

    Fetch the IEX minute intraday data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#historical-prices

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)
    use_date = work_dict.get('use_date', None)

    from_historical_date = None
    last_close_to_use = None
    dates = []

    if 'from_historical_date' in work_dict:
        from_historical_date = work_dict['from_historical_date']
        last_close_to_use = work_dict['last_close_to_use']
        dates = ae_utils.get_days_between_dates(
            from_historical_date=work_dict['from_historical_date'],
            last_close_to_use=last_close_to_use)

    use_url = (
        f'/stock/{ticker}/chart/1d')

    if use_date:
        use_url = (
            f'/stock/{ticker}/chart/date/{use_date}')

    log.debug(
        f'{label} - minute - url={use_url} '
        f'args={work_dict} ticker={ticker} '
        f'fhdate={from_historical_date} '
        f'last_close={last_close_to_use} '
        f'dates={dates}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame(resp_json)

    iex_helpers.convert_datetime_columns(
        df=df)

    if not use_date:
        use_date = df['date'].iloc[-1].strftime('%Y-%m-%d')

    new_minutes = dataset_utils.build_dates_from_df_col(
        src_col='minute',
        use_date_str=use_date,
        df=df)
    df['minute'] = pd.to_datetime(
        new_minutes,
        format=iex_consts.IEX_TICK_FORMAT)
    df.set_index(
        [
            'date',
            'minute'
        ])
    return df
# end of fetch_minute


def fetch_quote(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_quote

    Fetch the IEX quote data for a ticker and
    return as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#quote

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/quote')

    log.debug(
        f'{label} - quote - url={use_url} '
        f'args={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame([resp_json])

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_quote


def fetch_stats(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_stats

    Fetch the IEX statistics data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#key-stats

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/stats')

    log.debug(
        f'{label} - stats - url={use_url} '
        f'args={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame([resp_json])

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_stats


def fetch_peers(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_peers

    Fetch the IEX peers data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#peers

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/relevant')

    log.debug(
        f'{label} - peers - url={use_url} '
        f'args={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame(resp_json)

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_peers


def fetch_news(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_news

    Fetch the IEX news data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#news

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)
    num_news = int(work_dict.get('num_news', 5))

    use_url = (
        f'/stock/{ticker}/news/last/{num_news}')

    log.debug(
        f'{label} - news - url={use_url} '
        f'args={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame(resp_json)
    df['datetime'] = pd.to_datetime(
        df['datetime'],
        unit='ms',
        errors='coerce')

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_news


def fetch_financials(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_financials

    Fetch the IEX financial data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#financials

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/financials')

    log.debug(
        f'{label} - fins - url={use_url} '
        f'args={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame(resp_json.get('financials', []))

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_financials


def fetch_earnings(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_earnings

    Fetch the IEX earnings data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#earnings

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/earnings')

    log.debug(
        f'{label} - earns - url={use_url} '
        f'args={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame(resp_json.get('earnings', []))

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_earnings


def fetch_dividends(
        work_dict,
        scrub_mode='sort-by-date'):
    """fetch_dividends

    Fetch the IEX dividends data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#dividends

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)
    timeframe = work_dict.get('timeframe', '3m')

    use_url = (
        f'/stock/{ticker}/dividends/{timeframe}')

    log.debug(
        f'{label} - divs - url={use_url} '
        f'args={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame(resp_json)

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_dividends


def fetch_company(
        work_dict,
        scrub_mode='NO_SORT'):
    """fetch_company

    Fetch the IEX company data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#company

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    ticker = work_dict.get('ticker', None)
    label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/company')

    log.debug(
        f'{label} - comp - url={use_url} '
        f'args={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN)

    df = pd.DataFrame([resp_json])

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_company



"""
Helper for building an algorithm trading and performance history
as a dictionary that can be reviewed during or after an
algorithm finishes running
"""

import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def build_trade_history_entry(
        ticker,
        num_owned,
        close,
        balance,
        commission,
        date,
        trade_type,
        algo_start_price,
        original_balance,
        minute=None,
        high=None,
        low=None,
        open_val=None,
        volume=None,
        ask=None,
        bid=None,
        today_high=None,
        today_low=None,
        today_open_val=None,
        today_close=None,
        today_volume=None,
        stop_loss=None,
        trailing_stop_loss=None,
        buy_hold_units=None,
        sell_hold_units=None,
        spread_exp_date=None,
        spread_id=None,
        low_strike=None,
        low_bid=None,
        low_ask=None,
        low_volume=None,
        low_open_int=None,
        low_delta=None,
        low_gamma=None,
        low_theta=None,
        low_vega=None,
        low_rho=None,
        low_impl_vol=None,
        low_intrinsic=None,
        low_extrinsic=None,
        low_theo_price=None,
        low_theo_volatility=None,
        low_max_covered=None,
        low_exp_date=None,
        high_strike=None,
        high_bid=None,
        high_ask=None,
        high_volume=None,
        high_open_int=None,
        high_delta=None,
        high_gamma=None,
        high_theta=None,
        high_vega=None,
        high_rho=None,
        high_impl_vol=None,
        high_intrinsic=None,
        high_extrinsic=None,
        high_theo_price=None,
        high_theo_volatility=None,
        high_max_covered=None,
        high_exp_date=None,
        prev_balance=None,
        prev_num_owned=None,
        total_buys=None,
        total_sells=None,
        buy_triggered=None,
        buy_strength=None,
        buy_risk=None,
        sell_triggered=None,
        sell_strength=None,
        sell_risk=None,
        num_indicators_buy=None,
        num_indicators_sell=None,
        min_buy_indicators=None,
        min_sell_indicators=None,
        net_gain=None,
        net_value=None,
        ds_id=None,
        note=None,
        err=None,
        entry_spread_dict=None,
        version=1):
    """build_trade_history_entry

    Build a dictionary for tracking an algorithm profitability per ticker
    and for ``TRADE_SHARES``, ``TRADE_VERTICAL_BULL_SPREAD``, or
    ``TRADE_VERTICAL_BEAR_SPREAD`` trading types.

    .. note:: setting the ``minute`` is required to build
        a minute-by-minute ``Trading History``

    :param ticker: string ticker or symbol
    :param num_owned: integer current owned
        number of ``shares`` for this asset or number of
        currently owned ``contracts`` for an options
        spread.
    :param close: float ``close`` price of the
        underlying asset
    :param balance: float amount of available capital
    :param commission: float for commission costs
    :param date: string trade date for that row usually
        ``COMMON_DATE_FORMAT`` (``YYYY-MM-DD``)
    :param minute: optional - string for recording the minute
        the trade was place, and the format is
        ``COMMON_TICK_DATE_FORMAT`` (``YYYY-MM-DD HH:MM:SS``)
        this is optional if the algorithm is set up to
        trade using a ``day`` value for timeseries.
    :param trade_type: type of the trade - supported values:
            ``TRADE_SHARES``,
            ``TRADE_VERTICAL_BULL_SPREAD``,
            ``TRADE_VERTICAL_BEAR_SPREAD``
    :param algo_start_price: float starting close/contract price
        for this algo
    :param original_balance: float starting original account
        balance for this algo
    :param high: optional - float underlying stock asset ``high`` price
    :param low: optional - float underlying stock asset ``low`` price
    :param open_val: optional - float underlying stock asset ``open`` price
    :param volume: optional - integer underlying stock asset ``volume``
    :param ask: optional - float ``ask`` price of the
        stock (for buying ``shares``)
    :param bid: optional - float ``bid`` price of the
        stock (for selling ``shares``)
    :param today_high: optional - float ``high`` from
        the daily dataset (if available)
    :param today_low: optional - float ``low`` from
        the daily dataset (if available)
    :param today_open_val: optional - float ``open`` from
        the daily dataset (if available)
    :param today_close: optional - float ``close`` from
        the daily dataset (if available)
    :param today_volume: optional - float ``volume`` from
        the daily dataset (if available)
    :param stop_loss: optional - float ``stop_loss`` price of the
        stock/spread (for selling ``shares`` vs ``contracts``)
    :param trailing_stop_loss: optional - float ``trailing_stop_loss``
        price of the stock/spread (for selling ``shares`` vs ``contracts``)
    :param buy_hold_units: optional - number of units
        to hold buys - helps with algorithm tuning
    :param sell_hold_units: optional - number of units
        to hold sells - helps with algorithm tuning
    :param spread_exp_date: optional - string spread contract
        expiration date (``COMMON_DATE_FORMAT`` (``YYYY-MM-DD``)
    :param spread_id: optional - spread identifier for reviewing
        spread performances
    :param low_strike: optional
        - only for vertical bull/bear trade types
        ``low leg strike price`` of the spread
    :param low_bid: optional
        - only for vertical bull/bear trade types
        ``low leg bid`` of the spread
    :param low_ask: optional
        - only for vertical bull/bear trade types
        ``low leg ask`` of the spread
    :param low_volume: optional
        - only for vertical bull/bear trade types
        ``low leg volume`` of the spread
    :param low_open_int: optional
        - only for vertical bull/bear trade types
        ``low leg open interest`` of the spread
    :param low_delta: optional
        - only for vertical bull/bear trade types
        ``low leg delta`` of the spread
    :param low_gamma: optional
        - only for vertical bull/bear trade types
        ``low leg gamma`` of the spread
    :param low_theta: optional
        - only for vertical bull/bear trade types
        ``low leg theta`` of the spread
    :param low_vega: optional
        - only for vertical bull/bear trade types
        ``low leg vega`` of the spread
    :param low_rho: optional
        - only for vertical bull/bear trade types
        ``low leg rho`` of the spread
    :param low_impl_vol: optional
        - only for vertical bull/bear trade types
        ``low leg implied volatility`` of the spread
    :param low_intrinsic: optional
        - only for vertical bull/bear trade types
        ``low leg intrinsic`` of the spread
    :param low_extrinsic: optional
        - only for vertical bull/bear trade types
        ``low leg extrinsic`` of the spread
    :param low_theo_price: optional
        - only for vertical bull/bear trade types
        ``low leg theoretical price`` of the spread
    :param low_theo_volatility: optional
        - only for vertical bull/bear trade types
        ``low leg theoretical volatility`` of the spread
    :param low_max_covered: optional
        - only for vertical bull/bear trade types
        ``low leg max covered returns`` of the spread
    :param low_exp_date: optional
        - only for vertical bull/bear trade types
        ``low leg expiration date`` of the spread
    :param high_strike: optional
        - only for vertical bull/bear trade types
        ``high leg strike price`` of the spread
    :param high_bid: optional
        - only for vertical bull/bear trade types
        ``high leg bid`` of the spread
    :param high_ask: optional
        - only for vertical bull/bear trade types
        ``high leg ask`` of the spread
    :param high_volume: optional
        - only for vertical bull/bear trade types
        ``high leg volume`` of the spread
    :param high_open_int: optional
        - only for vertical bull/bear trade types
        ``high leg open interest`` of the spread
    :param high_delta: optional
        - only for vertical bull/bear trade types
        ``high leg delta`` of the spread
    :param high_gamma: optional
        - only for vertical bull/bear trade types
        ``high leg gamma`` of the spread
    :param high_theta: optional
        - only for vertical bull/bear trade types
        ``high leg theta`` of the spread
    :param high_vega: optional
        - only for vertical bull/bear trade types
        ``high leg vega`` of the spread
    :param high_rho: optional
        - only for vertical bull/bear trade types
        ``high leg rho`` of the spread
    :param high_impl_vol: optional
        - only for vertical bull/bear trade types
        ``high leg implied volatility`` of the spread
    :param high_intrinsic: optional
        - only for vertical bull/bear trade types
        ``high leg intrinsic`` of the spread
    :param high_extrinsic: optional
        - only for vertical bull/bear trade types
        ``high leg extrinsic`` of the spread
    :param high_theo_price: optional
        - only for vertical bull/bear trade types
        ``high leg theoretical price`` of the spread
    :param high_theo_volatility: optional
        - only for vertical bull/bear trade types
        ``high leg theoretical volatility`` of the spread
    :param high_max_covered: optional
        - only for vertical bull/bear trade types
        ``high leg max covered returns`` of the spread
    :param high_exp_date: optional
        - only for vertical bull/bear trade types
        ``high leg expiration date`` of the spread
    :param prev_balance: optional - previous balance
        for this algo
    :param prev_num_owned: optional - previous num of
        ``shares`` or ``contracts``
    :param total_buys: optional - total buy orders
        for this algo
    :param total_sells: optional - total sell orders
        for this algo
    :param buy_triggered: optional - bool
        ``buy`` conditions in the algorithm triggered
    :param buy_strength: optional - float
        custom strength/confidence rating for tuning
        algorithm performance for desirable
        sensitivity and specificity
    :param buy_risk: optional - float
        custom risk rating for tuning algorithm
        peformance for avoiding custom risk for buy
        conditions
    :param sell_triggered: optional - bool
        ``sell`` conditions in the algorithm triggered
    :param sell_strength: optional - float
        custom strength/confidence rating for tuning
        algorithm performance for desirable
        sensitivity and specificity
    :param sell_risk: optional - float
        custom risk rating for tuning algorithm
        peformance for avoiding custom risk for buy
        conditions
    :param num_indicators_buy: optional - integer
        number of indicators the ``IndicatorProcessor``
        processed and said to ``buy`` an asset
    :param num_indicators_sell: optional - integer
        number of indicators the ``IndicatorProcessor``
        processed and said to ``sell`` an asset
    :param min_buy_indicators: optional - integer
        minimum number of indicators required to trigger
        a ``buy`` order
    :param min_sell_indicators: optional - integer
        minimum number of indicators required to trigger
        a ``sell`` order
        net_gain=None,
        net_value=None,
    :param net_value: optional - float total value the algorithm
        has left remaining since starting trading. this includes
        the number of ``self.num_owned`` shares with the
        ``self.latest_close`` price included
    :param net_gain: optional - float amount the algorithm has
        made since starting including owned shares
        with the ``self.latest_close`` price included

    :param ds_id: optional - datset id for debugging
    :param note: optional - string for tracking high level
        testing notes on algorithm indicator ratings and
        internal message passing during an algorithms's
        ``self.process`` method
    :param err: optional - string for tracking errors
    :param entry_spread_dict: optional - on exit spreads
        the calculation of net gain can use the entry
        spread to determine specific performance metrics
        (work in progress)
    :param version: optional - version tracking order history
    """
    status = ae_consts.NOT_RUN
    algo_status = ae_consts.NOT_RUN
    err = None
    balance_net_gain = 0.0
    breakeven_price = None
    max_profit = None  # only for option spreads
    max_loss = None  # only for option spreads
    exp_date = None  # only for option spreads

    # latest price - start price of the algo
    price_change_since_start = close - algo_start_price

    if close:
        if close < 0.01:
            status = ae_consts.INVALID

    history_dict = {
        'ticker': ticker,
        'algo_start_price': ae_consts.to_f(algo_start_price),
        'algo_price_change': ae_consts.to_f(price_change_since_start),
        'original_balance': ae_consts.to_f(original_balance),
        'status': status,
        'algo_status': algo_status,
        'buy_now': buy_triggered,
        'buy_strength': buy_strength,
        'buy_risk': buy_risk,
        'sell_now': sell_triggered,
        'sell_strength': sell_strength,
        'sell_risk': sell_risk,
        'num_indicators_buy': num_indicators_buy,
        'num_indicators_sell': num_indicators_sell,
        'min_buy_indicators': min_buy_indicators,
        'min_sell_indicators': min_sell_indicators,
        'ds_id': ds_id,
        'num_owned': num_owned,
        'close': ae_consts.to_f(close),
        'balance': ae_consts.to_f(balance),
        'commission': ae_consts.to_f(commission),
        'date': date,
        'minute': minute,
        'trade_type': trade_type,
        'high': ae_consts.to_f(high),
        'low': ae_consts.to_f(low),
        'open': ae_consts.to_f(open_val),
        'volume': volume,
        'ask': ae_consts.to_f(ask),
        'bid': ae_consts.to_f(bid),
        'today_high': ae_consts.to_f(today_high),
        'today_low': ae_consts.to_f(today_low),
        'today_open_val': ae_consts.to_f(today_open_val),
        'today_close': ae_consts.to_f(today_close),
        'today_volume': ae_consts.to_f(today_volume),
        'stop_loss': ae_consts.to_f(stop_loss),
        'trailing_stop_loss': ae_consts.to_f(trailing_stop_loss),
        'buy_hold_units': buy_hold_units,
        'sell_hold_units': sell_hold_units,
        'low_strike': low_strike,
        'low_bid': ae_consts.to_f(low_bid),
        'low_ask': ae_consts.to_f(low_ask),
        'low_volume': low_volume,
        'low_open_int': low_open_int,
        'low_delta': low_delta,
        'low_gamma': low_gamma,
        'low_theta': low_theta,
        'low_vega': low_vega,
        'low_rho': low_rho,
        'low_impl_vol': low_impl_vol,
        'low_intrinsic': low_intrinsic,
        'low_extrinsic': low_extrinsic,
        'low_theo_price': low_theo_price,
        'low_theo_volatility': low_theo_volatility,
        'low_max_covered': low_max_covered,
        'low_exp_date': low_exp_date,
        'high_strike': high_strike,
        'high_bid': ae_consts.to_f(high_bid),
        'high_ask': ae_consts.to_f(high_ask),
        'high_volume': high_volume,
        'high_open_int': high_open_int,
        'high_delta': high_delta,
        'high_gamma': high_gamma,
        'high_theta': high_theta,
        'high_vega': high_vega,
        'high_rho': high_rho,
        'high_impl_vol': high_impl_vol,
        'high_intrinsic': high_intrinsic,
        'high_extrinsic': high_extrinsic,
        'high_theo_price': high_theo_price,
        'high_theo_volatility': high_theo_volatility,
        'high_max_covered': high_max_covered,
        'high_exp_date': high_exp_date,
        'spread_id': spread_id,
        'net_gain': net_gain,
        'net_value': net_value,
        'breakeven_price': breakeven_price,
        'max_profit': ae_consts.to_f(max_profit),
        'max_loss': ae_consts.to_f(max_loss),
        'exp_date': exp_date,
        'prev_balance': ae_consts.to_f(prev_balance),
        'prev_num_owned': ae_consts.to_f(prev_num_owned),
        'total_buys': total_buys,
        'total_sells': total_sells,
        'note': note,
        'err': err,
        'version': version
    }

    # evaluate if the algorithm is gaining
    # cash over the test
    if balance and original_balance:
        # net change on the balance
        # note this needs to be upgraded to
        # support orders per ticker
        # single tickers will work for v1
        balance_net_gain = balance - original_balance
        if balance_net_gain > 0.0:
            algo_status = ae_consts.ALGO_PROFITABLE
        else:
            algo_status = ae_consts.ALGO_NOT_PROFITABLE
    else:
        history_dict['err'] = (
            f'{ticker} ds_id={ds_id} missing balance={balance} and '
            f'original_balance={original_balance}')
        algo_status = ae_consts.ALGO_ERROR
    # if starting balance and original_balance exist
    # to determine algorithm trade profitability

    # if there are no shares to sell then
    # there's no current trade open
    if num_owned and num_owned < 1:
        status = ae_consts.TRADE_NO_SHARES_TO_SELL
    else:
        if close < 0.01:
            history_dict['err'] = (
                f'{ticker} ds_id={ds_id} close={close} must be greater '
                f'than 0.01')
            status = ae_consts.TRADE_ERROR
        elif algo_start_price < 0.01:
            history_dict['err'] = (
                f'{ticker} ds_id={ds_id} '
                f'algo_start_price={algo_start_price} must be greater '
                f'than 0.01')
            status = ae_consts.TRADE_ERROR
        else:
            price_net_gain = close - algo_start_price
            if price_net_gain > 0.0:
                status = ae_consts.TRADE_PROFITABLE
            else:
                status = ae_consts.TRADE_NOT_PROFITABLE
    # if starting price when algo started and close exist
    # determine if this trade profitability

    # Assign calculated values:
    history_dict['net_gain'] = net_gain
    history_dict['balance_net_gain'] = balance_net_gain
    history_dict['breakeven_price'] = breakeven_price
    history_dict['max_profit'] = max_profit
    history_dict['max_loss'] = max_loss
    history_dict['exp_date'] = exp_date

    # assign statuses
    history_dict['status'] = status
    history_dict['algo_status'] = algo_status

    use_date = minute
    if not use_date:
        use_date = date

    log.debug(
        f'{ticker} ds_id={ds_id} {use_date} '
        f'algo={ae_consts.get_status(status=history_dict["algo_status"])} '
        f'trade={ae_consts.get_status(status=history_dict["status"])} '
        f'history={ae_consts.ppj(history_dict)}')

    return history_dict
# end of build_trade_history_entry
